@@ -47,12 +47,24 @@ struct NanValue {
     static inline int32_t NO_PATH() { return -2; }
 };
 
+// Musl doesn't have nanf implementation we need, just check if they are the
+// same value. If not, prefer old hardcoded value.
+bool isGoodNanf() {
+    int32_t noValue = decodeValue(std::nanf("1"));
+    int32_t noPath = decodeValue(std::nanf("2"));
+    return (noValue != noPath);
+}
+
 template <>
 struct NanValue<float> {
     static_assert(std::numeric_limits<float>::has_quiet_NaN,
                   "Require support for quiet NaN.");
-    static inline int32_t NO_VALUE() { return decodeValue(std::nanf("1")); }
-    static inline int32_t NO_PATH() { return decodeValue(std::nanf("2")); }
+    static inline int32_t NO_VALUE() {
+        return isGoodNanf() ? decodeValue(std::nanf("1")) : 0x7fc00001;
+    }
+    static inline int32_t NO_PATH() {
+        return isGoodNanf() ? decodeValue(std::nanf("2")) : 0x7fc00002;
+    }
 };
 
 } // namespace
@@ -1100,6 +1112,11 @@ typename DATrie<T>::value_type DATrie<T>::exactMatchSearch(const char *key,
     return decoder.result_value;
 }
 
+template <typename T>
+bool DATrie<T>::hasExactMatch(std::string_view key) const {
+    return isValid(exactMatchSearch(key));
+}
+
 template <typename T>
 typename DATrie<T>::value_type DATrie<T>::traverse(const char *key, size_t len,
                                                    position_type &from) const {
@@ -1139,6 +1156,20 @@ bool DATrie<T>::isValid(value_type v) {
     return !(isNoPath(v) || isNoValue(v));
 }
 
+template <typename T>
+T DATrie<T>::noPath() {
+    typename DATriePrivate<T>::decorder_type d;
+    d.result = DATriePrivate<value_type>::CEDAR_NO_PATH;
+    return d.result_value;
+}
+
+template <typename T>
+T DATrie<T>::noValue() {
+    typename DATriePrivate<T>::decorder_type d;
+    d.result = DATriePrivate<value_type>::CEDAR_NO_VALUE;
+    return d.result_value;
+}
+
 template <typename T>
 size_t DATrie<T>::mem_size() const {
     //     std::cout << "tail" << d->m_tail.size() << std::endl
