@@ -5,7 +5,12 @@
  */
 
 #include "pinyindata.h"
+#include "libime/pinyin/pinyinencoder.h"
+#include <boost/algorithm/string/predicate.hpp>
+#include <fcitx-utils/stringutils.h>
+#include <unordered_map>
 #include <unordered_set>
+#include <utility>
 
 namespace libime {
 
@@ -131,6 +136,15 @@ getInnerSegment() {
     return innerSegment;
 }
 
+inline bool operator==(const PinyinEntry &a, const PinyinEntry &b) {
+    return a.pinyin() == b.pinyin() && a.initial() == b.initial() &&
+           a.final() == b.final() && a.flags() == b.flags();
+}
+
+inline bool operator!=(const PinyinEntry &a, const PinyinEntry &b) {
+    return !(a == b);
+}
+
 const PinyinMap &getPinyinMap() {
     static const PinyinMap pinyinMap = {
         {"zuo", PinyinInitial::Z, PinyinFinal::UO, PinyinFuzzyFlag::None},
@@ -154,6 +168,8 @@ const PinyinMap &getPinyinMap() {
         {"zou", PinyinInitial::Z, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"zogn", PinyinInitial::Z, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"zon", PinyinInitial::Z, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"zong", PinyinInitial::Z, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"zi", PinyinInitial::Z, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"zhuo", PinyinInitial::ZH, PinyinFinal::UO, PinyinFuzzyFlag::None},
@@ -169,6 +185,8 @@ const PinyinMap &getPinyinMap() {
         {"zhou", PinyinInitial::ZH, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"zhogn", PinyinInitial::ZH, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"zhon", PinyinInitial::ZH, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"zhong", PinyinInitial::ZH, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"zhi", PinyinInitial::ZH, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"zhegn", PinyinInitial::ZH, PinyinFinal::ENG,
@@ -222,6 +240,8 @@ const PinyinMap &getPinyinMap() {
         {"you", PinyinInitial::Y, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"yogn", PinyinInitial::Y, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"yon", PinyinInitial::Y, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"yong", PinyinInitial::Y, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"yo", PinyinInitial::Y, PinyinFinal::O, PinyinFuzzyFlag::None},
         {"yign", PinyinInitial::Y, PinyinFinal::ING,
@@ -304,6 +324,8 @@ const PinyinMap &getPinyinMap() {
         {"tou", PinyinInitial::T, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"togn", PinyinInitial::T, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"ton", PinyinInitial::T, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"tong", PinyinInitial::T, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"tign", PinyinInitial::T, PinyinFinal::ING,
          PinyinFuzzyFlag::CommonTypo},
@@ -353,6 +375,8 @@ const PinyinMap &getPinyinMap() {
         {"sou", PinyinInitial::S, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"sogn", PinyinInitial::S, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"son", PinyinInitial::S, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"song", PinyinInitial::S, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"si", PinyinInitial::S, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"shuo", PinyinInitial::SH, PinyinFinal::UO, PinyinFuzzyFlag::None},
@@ -408,6 +432,8 @@ const PinyinMap &getPinyinMap() {
         {"rou", PinyinInitial::R, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"rogn", PinyinInitial::R, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"ron", PinyinInitial::R, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"rong", PinyinInitial::R, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"ri", PinyinInitial::R, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"regn", PinyinInitial::R, PinyinFinal::ENG,
@@ -507,6 +533,8 @@ const PinyinMap &getPinyinMap() {
         {"nou", PinyinInitial::N, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"nogn", PinyinInitial::N, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"non", PinyinInitial::N, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"nong", PinyinInitial::N, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"niu", PinyinInitial::N, PinyinFinal::IU, PinyinFuzzyFlag::None},
         {"nign", PinyinInitial::N, PinyinFinal::ING,
@@ -584,6 +612,8 @@ const PinyinMap &getPinyinMap() {
         {"lou", PinyinInitial::L, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"logn", PinyinInitial::L, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"lon", PinyinInitial::L, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"long", PinyinInitial::L, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"lo", PinyinInitial::L, PinyinFinal::O, PinyinFuzzyFlag::None},
         {"liu", PinyinInitial::L, PinyinFinal::IU, PinyinFuzzyFlag::None},
@@ -625,6 +655,8 @@ const PinyinMap &getPinyinMap() {
         {"kou", PinyinInitial::K, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"kogn", PinyinInitial::K, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"kon", PinyinInitial::K, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"kong", PinyinInitial::K, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"kegn", PinyinInitial::K, PinyinFinal::ENG,
          PinyinFuzzyFlag::CommonTypo},
@@ -692,6 +724,8 @@ const PinyinMap &getPinyinMap() {
         {"hou", PinyinInitial::H, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"hogn", PinyinInitial::H, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"hon", PinyinInitial::H, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"hong", PinyinInitial::H, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"hegn", PinyinInitial::H, PinyinFinal::ENG,
          PinyinFuzzyFlag::CommonTypo},
@@ -719,6 +753,8 @@ const PinyinMap &getPinyinMap() {
         {"gou", PinyinInitial::G, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"gogn", PinyinInitial::G, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"gon", PinyinInitial::G, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"gong", PinyinInitial::G, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"gegn", PinyinInitial::G, PinyinFinal::ENG,
          PinyinFuzzyFlag::CommonTypo},
@@ -769,6 +805,8 @@ const PinyinMap &getPinyinMap() {
         {"dou", PinyinInitial::D, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"dogn", PinyinInitial::D, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"don", PinyinInitial::D, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"dong", PinyinInitial::D, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"diu", PinyinInitial::D, PinyinFinal::IU, PinyinFuzzyFlag::None},
         {"dign", PinyinInitial::D, PinyinFinal::ING,
@@ -819,6 +857,8 @@ const PinyinMap &getPinyinMap() {
         {"cou", PinyinInitial::C, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"cogn", PinyinInitial::C, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"con", PinyinInitial::C, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"cong", PinyinInitial::C, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"ci", PinyinInitial::C, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"chuo", PinyinInitial::CH, PinyinFinal::UO, PinyinFuzzyFlag::None},
@@ -834,6 +874,8 @@ const PinyinMap &getPinyinMap() {
         {"chou", PinyinInitial::CH, PinyinFinal::OU, PinyinFuzzyFlag::None},
         {"chogn", PinyinInitial::CH, PinyinFinal::ONG,
          PinyinFuzzyFlag::CommonTypo},
+        {"chon", PinyinInitial::CH, PinyinFinal::ONG,
+         PinyinFuzzyFlag::CommonTypo},
         {"chong", PinyinInitial::CH, PinyinFinal::ONG, PinyinFuzzyFlag::None},
         {"chi", PinyinInitial::CH, PinyinFinal::I, PinyinFuzzyFlag::None},
         {"chegn", PinyinInitial::CH, PinyinFinal::ENG,
@@ -897,4 +939,322 @@ const PinyinMap &getPinyinMap() {
     };
     return pinyinMap;
 }
+
+std::optional<PinyinEntry> applyFuzzy(const PinyinEntry &entry,
+                                      PinyinFuzzyFlag fz, int pass) {
+    if (entry.pinyin() == "m" || entry.pinyin() == "n" ||
+        entry.pinyin() == "r" || entry.pinyin() == "ng" ||
+        entry.pinyin() == "ou") {
+        return std::nullopt;
+    }
+    auto result = entry.pinyin();
+    switch (fz) {
+    case PinyinFuzzyFlag::CommonTypo: {
+        if (pass == 0) {
+            // Allow non standard usage like jv jve jvan jvuang
+            if (result[0] == 'j' || result[0] == 'q' || result[0] == 'x' ||
+                result[0] == 'y') {
+                if (boost::algorithm::ends_with(result, "u") &&
+                    !boost::algorithm::ends_with(result, "iu") &&
+                    !boost::algorithm::ends_with(result, "ou")) {
+                    result.back() = 'v';
+                }
+
+                if (boost::algorithm::ends_with(result, "ue")) {
+                    result[result.size() - 2] = 'v';
+                }
+                if (boost::algorithm::ends_with(result, "uan")) {
+                    result[result.size() - 3] = 'v';
+                }
+                if (boost::algorithm::ends_with(result, "uang")) {
+                    result[result.size() - 4] = 'v';
+                }
+            }
+        } else if (pass == 1) {
+            // Allow lon -> long
+            if (boost::algorithm::ends_with(result, "ong")) {
+                result.pop_back();
+            }
+        } else if (pass == 2) {
+            // Allow ying -> yign
+            if (boost::algorithm::ends_with(result, "ng")) {
+                result[result.size() - 2] = 'g';
+                result[result.size() - 1] = 'n';
+            } else if (boost::algorithm::ends_with(result, "ue")) {
+                // Allow fuzzy for uv, that does not cause ambiguity.
+                result[result.size() - 2] = 'e';
+                result[result.size() - 1] = 'u';
+            } else if (boost::algorithm::ends_with(result, "ve")) {
+                result[result.size() - 2] = 'e';
+                result[result.size() - 1] = 'v';
+            } else if (boost::algorithm::ends_with(result, "ua")) {
+                result[result.size() - 2] = 'a';
+                result[result.size() - 1] = 'u';
+            } else if (boost::algorithm::ends_with(result, "uai")) {
+                result[result.size() - 3] = 'a';
+                result[result.size() - 2] = 'u';
+            } else if (boost::algorithm::ends_with(result, "uan")) {
+                result[result.size() - 3] = 'a';
+                result[result.size() - 2] = 'u';
+            } else if (boost::algorithm::ends_with(result, "van")) {
+                result[result.size() - 3] = 'a';
+                result[result.size() - 2] = 'v';
+            }
+        } else if (pass == 3) {
+            // this conflicts with "ng" rule, so need a separate pass.
+            if (boost::algorithm::ends_with(result, "uang")) {
+                result[result.size() - 4] = 'a';
+                result[result.size() - 3] = 'u';
+            } else if (boost::algorithm::ends_with(result, "vang")) {
+                result[result.size() - 4] = 'a';
+                result[result.size() - 3] = 'v';
+            }
+        }
+        break;
+    case PinyinFuzzyFlag::AdvancedTypo:
+        if (pass == 0) {
+            // Allow reversed zhe -> hze
+            if (boost::algorithm::starts_with(result, "zh") ||
+                boost::algorithm::starts_with(result, "sh") ||
+                boost::algorithm::starts_with(result, "ch")) {
+                std::swap(result[0], result[1]);
+            } else if (boost::algorithm::ends_with(result, "un") &&
+                       !boost::algorithm::ends_with(result, "aun")) {
+                result[result.size() - 2] = 'n';
+                result[result.size() - 1] = 'u';
+            }
+        } else if (pass == 1) {
+            if (entry.flags().test(PinyinFuzzyFlag::AdvancedTypo)) {
+                break;
+            }
+            for (const auto two : {"ai", "ia", "ei", "ie", "ao", "uo", "ou",
+                                   "iu", "an", "en", "in"}) {
+                if (boost::algorithm::ends_with(result, two)) {
+                    std::swap(result[result.size() - 2],
+                              result[result.size() - 1]);
+                }
+            }
+        } else if (pass == 2) {
+            if (entry.flags().test(PinyinFuzzyFlag::AdvancedTypo)) {
+                break;
+            }
+            for (const auto three :
+                 {"ang", "eng", "ing", "ong", "iao", "ian"}) {
+                if (boost::algorithm::ends_with(result, three)) {
+                    std::swap(result[result.size() - 3],
+                              result[result.size() - 2]);
+                }
+            }
+        } else if (pass == 3) {
+            if (entry.flags().test(PinyinFuzzyFlag::AdvancedTypo)) {
+                break;
+            }
+            for (const auto four : {"iang", "iong"}) {
+                if (boost::algorithm::ends_with(result, four)) {
+                    std::swap(result[result.size() - 4],
+                              result[result.size() - 3]);
+                }
+            }
+        } else if (pass == 4) {
+            if (entry.flags().test(PinyinFuzzyFlag::AdvancedTypo)) {
+                break;
+            }
+            // zhe -> zeh.
+            if (result.size() == 3 && result[1] == 'h' &&
+                entry.flags() == PinyinFuzzyFlag::None) {
+                std::swap(result[result.size() - 2], result[result.size() - 1]);
+            }
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::VE_UE: {
+        if (boost::algorithm::ends_with(result, "ve")) {
+            result[result.size() - 2] = 'u';
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::IAN_IANG: {
+        if (boost::algorithm::ends_with(result, "ian")) {
+            result.push_back('g');
+        } else if (boost::algorithm::ends_with(result, "iang")) {
+            result.pop_back();
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::UAN_UANG: {
+        if (entry.flags() != PinyinFuzzyFlag::None) {
+            break;
+        }
+        if (boost::algorithm::ends_with(result, "uan")) {
+            result.push_back('g');
+        } else if (boost::algorithm::ends_with(result, "uang")) {
+            result.pop_back();
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::AN_ANG: {
+        if (boost::algorithm::ends_with(result, "uan") ||
+            boost::algorithm::ends_with(result, "uang")) {
+            break;
+        }
+        if (boost::algorithm::ends_with(result, "ian") ||
+            boost::algorithm::ends_with(result, "iang")) {
+            break;
+        }
+        if (boost::algorithm::ends_with(result, "an")) {
+            result.push_back('g');
+        } else if (boost::algorithm::ends_with(result, "ang")) {
+            result.pop_back();
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::EN_ENG: {
+        if (boost::algorithm::ends_with(result, "en")) {
+            result.push_back('g');
+        } else if (boost::algorithm::ends_with(result, "eng")) {
+            result.pop_back();
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::IN_ING: {
+        if (boost::algorithm::ends_with(result, "in")) {
+            result.push_back('g');
+        } else if (boost::algorithm::ends_with(result, "ing")) {
+            result.pop_back();
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::U_OU: {
+        if (boost::algorithm::ends_with(result, "ou")) {
+            result.pop_back();
+            result.back() = 'u';
+        } else if (boost::algorithm::ends_with(result, "u") &&
+                   !boost::algorithm::ends_with(result, "iu")) {
+            result.back() = 'o';
+            result.push_back('u');
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::C_CH: {
+        if (entry.flags() != PinyinFuzzyFlag::None) {
+            break;
+        }
+        if (boost::algorithm::starts_with(result, "ch")) {
+            result.erase(std::next(result.begin()));
+        } else if (boost::algorithm::starts_with(result, "c")) {
+            result.insert(std::next(result.begin()), 'h');
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::S_SH: {
+        if (entry.flags() != PinyinFuzzyFlag::None) {
+            break;
+        }
+        if (boost::algorithm::starts_with(result, "sh")) {
+            result.erase(std::next(result.begin()));
+        } else if (boost::algorithm::starts_with(result, "s")) {
+            result.insert(std::next(result.begin()), 'h');
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::Z_ZH: {
+        if (entry.flags() != PinyinFuzzyFlag::None) {
+            break;
+        }
+        if (boost::algorithm::starts_with(result, "zh")) {
+            result.erase(std::next(result.begin()));
+        } else if (boost::algorithm::starts_with(result, "z")) {
+            result.insert(std::next(result.begin()), 'h');
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::F_H: {
+        if (boost::algorithm::starts_with(result, "f")) {
+            result.front() = 'h';
+        } else if (boost::algorithm::starts_with(result, "h")) {
+            result.front() = 'f';
+        }
+        break;
+    }
+
+    case PinyinFuzzyFlag::L_N: {
+        if (boost::algorithm::starts_with(result, "l")) {
+            result.front() = 'n';
+        } else if (boost::algorithm::starts_with(result, "n")) {
+            result.front() = 'l';
+        }
+        break;
+    }
+    default:
+        break;
+    }
+    if (result == entry.pinyin()) {
+        return std::nullopt;
+    }
+    return PinyinEntry(result.data(), entry.initial(), entry.final(),
+                       entry.flags() | fz);
+}
+
+void applyFuzzy(PinyinMap &map, PinyinFuzzyFlag fz, int pass = 0) {
+    std::vector<PinyinEntry> newEntries;
+    for (auto &entry : map) {
+        if (auto newEntry = applyFuzzy(entry, fz, pass)) {
+            newEntries.push_back(*newEntry);
+        }
+    }
+
+    for (const auto &newEntry : newEntries) {
+        if (auto iter = map.find(newEntry.pinyin()); iter != map.end()) {
+            if (iter->flags() == PinyinFuzzyFlag::None) {
+                continue;
+            }
+        }
+        FCITX_ASSERT(map.insert(newEntry).second);
+    }
+}
+
+const PinyinMap &getPinyinMapV2() {
+    static const PinyinMap map = []() {
+        const PinyinMap &orig = getPinyinMap();
+
+        PinyinMap filtered;
+        for (const auto &entry : orig) {
+            if (entry.flags() == PinyinFuzzyFlag::None) {
+                filtered.insert(entry);
+            }
+        }
+
+        for (auto fz : {PinyinFuzzyFlag::U_OU, PinyinFuzzyFlag::IN_ING,
+                        PinyinFuzzyFlag::EN_ENG, PinyinFuzzyFlag::AN_ANG,
+                        PinyinFuzzyFlag::UAN_UANG, PinyinFuzzyFlag::IAN_IANG,
+                        PinyinFuzzyFlag::VE_UE, PinyinFuzzyFlag::F_H,
+                        PinyinFuzzyFlag::L_N, PinyinFuzzyFlag::Z_ZH,
+                        PinyinFuzzyFlag::S_SH, PinyinFuzzyFlag::C_CH}) {
+            applyFuzzy(filtered, fz);
+        }
+        applyFuzzy(filtered, PinyinFuzzyFlag::CommonTypo, 0);
+        applyFuzzy(filtered, PinyinFuzzyFlag::CommonTypo, 1);
+        applyFuzzy(filtered, PinyinFuzzyFlag::CommonTypo, 2);
+        applyFuzzy(filtered, PinyinFuzzyFlag::CommonTypo, 3);
+        applyFuzzy(filtered, PinyinFuzzyFlag::AdvancedTypo, 0);
+        applyFuzzy(filtered, PinyinFuzzyFlag::AdvancedTypo, 1);
+        applyFuzzy(filtered, PinyinFuzzyFlag::AdvancedTypo, 2);
+        applyFuzzy(filtered, PinyinFuzzyFlag::AdvancedTypo, 3);
+        applyFuzzy(filtered, PinyinFuzzyFlag::AdvancedTypo, 4);
+        return filtered;
+    }();
+    return map;
+}
+
 } // namespace libime
