@@ -21,15 +21,17 @@
 #include <fstream>
 #include <iomanip>
 #include <queue>
+#include <string>
 #include <string_view>
 #include <type_traits>
 
 namespace libime {
 
-static const float fuzzyCost = std::log10(0.5f);
-static const size_t minimumLongWordLength = 3;
-static const float invalidPinyinCost = -100.0f;
-static const char pinyinHanziSep = '!';
+namespace {
+const float fuzzyCost = std::log10(0.5f);
+const size_t minimumLongWordLength = 3;
+const float invalidPinyinCost = -100.0f;
+const char pinyinHanziSep = '!';
 
 static constexpr uint32_t pinyinBinaryFormatMagic = 0x000fc613;
 static constexpr uint32_t pinyinBinaryFormatVersion = 0x2;
@@ -148,6 +150,47 @@ const SegmentGraphNode *prevIsSeparator(const SegmentGraph &graph,
     return nullptr;
 }
 
+inline void searchOneStep(
+    std::list<std::pair<const PinyinTrie *, PinyinTrie::position_type>> &nodes,
+    char current) {
+    std::list<std::pair<const PinyinTrie *, PinyinTrie::position_type>>
+        extraNodes;
+    auto iter = nodes.begin();
+    while (iter != nodes.end()) {
+        if (current != 0) {
+            PinyinTrie::value_type result;
+            result = iter->first->traverse(&current, 1, iter->second);
+
+            if (PinyinTrie::isNoPath(result)) {
+                nodes.erase(iter++);
+            } else {
+                iter++;
+            }
+        } else {
+            bool changed = false;
+            for (char test = PinyinEncoder::firstFinal;
+                 test <= PinyinEncoder::lastFinal; test++) {
+                decltype(extraNodes)::value_type p = *iter;
+                auto result = p.first->traverse(&test, 1, p.second);
+                if (!PinyinTrie::isNoPath(result)) {
+                    extraNodes.push_back(p);
+                    changed = true;
+                }
+            }
+            if (changed) {
+                *iter = extraNodes.back();
+                extraNodes.pop_back();
+                iter++;
+            } else {
+                nodes.erase(iter++);
+            }
+        }
+    }
+    nodes.splice(nodes.end(), std::move(extraNodes));
+}
+
+} // namespace
+
 class PinyinMatchContext {
 public:
     explicit PinyinMatchContext(
@@ -598,39 +641,7 @@ void PinyinDictionary::matchWords(const char *data, size_t size,
         } else {
             current = pinyinHanziSep;
         }
-        decltype(nodes) extraNodes;
-        auto iter = nodes.begin();
-        while (iter != nodes.end()) {
-            if (current != 0) {
-                PinyinTrie::value_type result;
-                result = iter->first->traverse(&current, 1, iter->second);
-
-                if (PinyinTrie::isNoPath(result)) {
-                    nodes.erase(iter++);
-                } else {
-                    iter++;
-                }
-            } else {
-                bool changed = false;
-                for (char test = PinyinEncoder::firstFinal;
-                     test <= PinyinEncoder::lastFinal; test++) {
-                    decltype(extraNodes)::value_type p = *iter;
-                    auto result = p.first->traverse(&test, 1, p.second);
-                    if (!PinyinTrie::isNoPath(result)) {
-                        extraNodes.push_back(p);
-                        changed = true;
-                    }
-                }
-                if (changed) {
-                    *iter = extraNodes.back();
-                    extraNodes.pop_back();
-                    iter++;
-                } else {
-                    nodes.erase(iter++);
-                }
-            }
-        }
-        nodes.splice(nodes.end(), std::move(extraNodes));
+        searchOneStep(nodes, current);
     }
 
     for (auto &node : nodes) {
@@ -641,12 +652,51 @@ void PinyinDictionary::matchWords(const char *data, size_t size,
                 node.first->suffix(s, len + size + 1, pos);
 
                 auto view = std::string_view(s);
-                return callback(s.substr(0, size), view.substr(size + 1),
+                return callback(view.substr(0, size), view.substr(size + 1),
                                 value);
             },
             node.second);
     }
 }
+
+void PinyinDictionary::matchWordsPrefix(const char *data, size_t size,
+                                        PinyinMatchCallback callback) const {
+    if (!PinyinEncoder::isValidUserPinyin(data, size)) {
+        return;
+    }
+
+    FCITX_D();
+    std::list<std::pair<const PinyinTrie *, PinyinTrie::position_type>> nodes;
+    for (size_t i = 0; i < dictSize(); i++) {
+        if (d->flags_[i].test(PinyinDictFlag::Disabled)) {
+            continue;
+        }
+        const auto &trie = *this->trie(i);
+        nodes.emplace_back(&trie, 0);
+    }
+    for (size_t i = 0; i < size && !nodes.empty(); i++) {
+        searchOneStep(nodes, data[i]);
+    }
+
+    for (auto &node : nodes) {
+        node.first->foreach(
+            [&node, &callback, size](PinyinTrie::value_type value, size_t len,
+                                     uint64_t pos) {
+                std::string s;
+                node.first->suffix(s, len + size, pos);
+
+                std::string_view view(s);
+                if (auto sep = view.find(pinyinHanziSep, size);
+                    sep != std::string::npos) {
+                    return callback(view.substr(0, sep), view.substr(sep + 1),
+                                    value);
+                }
+                return true;
+            },
+            node.second);
+    }
+}
+
 PinyinDictionary::PinyinDictionary()
     : d_ptr(std::make_unique<PinyinDictionaryPrivate>(this)) {
     FCITX_D();
@@ -730,21 +780,10 @@ void PinyinDictionary::loadBinary(size_t idx, std::istream &in) {
     case 0x1:
         trie.load(in);
         break;
-    case pinyinBinaryFormatVersion: {
-        boost::iostreams::filtering_istreambuf compressBuf;
-        compressBuf.push(ZSTDDecompressor());
-        compressBuf.push(in);
-        std::istream compressIn(&compressBuf);
-
-        trie.load(compressIn);
-        // We don't want to read any data, but only trigger the zstd footer
-        // handling, which validates CRC.
-        compressIn.peek();
-        if (compressIn.bad()) {
-            throw std::invalid_argument("Failed to load dict data");
-        }
+    case pinyinBinaryFormatVersion:
+        readZSTDCompressed(
+            in, [&trie](std::istream &compressIn) { trie.load(compressIn); });
         break;
-    }
     default:
         throw std::invalid_argument("Invalid pinyin version.");
         break;
@@ -769,12 +808,9 @@ void PinyinDictionary::save(size_t idx, std::ostream &out,
         throw_if_io_fail(marshall(out, pinyinBinaryFormatMagic));
         throw_if_io_fail(marshall(out, pinyinBinaryFormatVersion));
 
-        boost::iostreams::filtering_streambuf<boost::iostreams::output>
-            compressBuf;
-        compressBuf.push(ZSTDCompressor());
-        compressBuf.push(out);
-        std::ostream compressOut(&compressBuf);
-        mutableTrie(idx)->save(compressOut);
+        writeZSTDCompressed(out, [this, idx](std::ostream &compressOut) {
+            mutableTrie(idx)->save(compressOut);
+        });
     } break;
     default:
         throw std::invalid_argument("invalid format type");
