@@ -35,7 +35,8 @@ size_t sentenceCodeLength(const SentenceResult &sentence) {
 template <OrderPolicy policy>
 struct TableCandidateCompare {
     TableCandidateCompare(int noSortInputLength, bool sortByCodeLength)
-        : noSortInputLength_(noSortInputLength), sortByCodeLength_(sortByCodeLength) {}
+        : noSortInputLength_(noSortInputLength),
+          sortByCodeLength_(sortByCodeLength) {}
 
     // Larger index should be put ahead.
     static int64_t index(const SentenceResult &sentence) {
@@ -172,7 +173,17 @@ public:
         if (!canDoAutoSelect()) {
             return false;
         }
-        return candidates_.size() == 1;
+        if (candidates_.size() != 1) {
+            return false;
+        }
+
+        if (candidates_[0].sentence().size() != 1) {
+            return false;
+        }
+        FCITX_Q();
+        return q->code(candidates_[0]) == q->currentCode() &&
+               (!dict_.tableOptions().exactMatch() ||
+                dict_.hasOneMatchingWord(q->currentCode()));
     };
 
     State currentState() {
@@ -562,16 +573,22 @@ void TableContext::update() {
 
         switch (d->dict_.tableOptions().orderPolicy()) {
         case OrderPolicy::No:
-            std::sort(d->candidates_.begin(), d->candidates_.end(),
-                      TableCandidateCompare<OrderPolicy::No>(noSortLength, d->dict_.tableOptions().sortByCodeLength()));
+            std::sort(
+                d->candidates_.begin(), d->candidates_.end(),
+                TableCandidateCompare<OrderPolicy::No>(
+                    noSortLength, d->dict_.tableOptions().sortByCodeLength()));
             break;
         case OrderPolicy::Fast:
-            std::sort(d->candidates_.begin(), d->candidates_.end(),
-                      TableCandidateCompare<OrderPolicy::Fast>(noSortLength, d->dict_.tableOptions().sortByCodeLength()));
+            std::sort(
+                d->candidates_.begin(), d->candidates_.end(),
+                TableCandidateCompare<OrderPolicy::Fast>(
+                    noSortLength, d->dict_.tableOptions().sortByCodeLength()));
             break;
         case OrderPolicy::Freq:
-            std::sort(d->candidates_.begin(), d->candidates_.end(),
-                      TableCandidateCompare<OrderPolicy::Freq>(noSortLength, d->dict_.tableOptions().sortByCodeLength()));
+            std::sort(
+                d->candidates_.begin(), d->candidates_.end(),
+                TableCandidateCompare<OrderPolicy::Freq>(
+                    noSortLength, d->dict_.tableOptions().sortByCodeLength()));
             break;
         }
         if (!d->candidates_.empty() && isPinyin(d->candidates_[0])) {
