@@ -4,6 +4,7 @@
  * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 #include "libime/pinyin/pinyindata.h"
+#include "libime/pinyin/pinyinencoder.h"
 #include "libime/pinyin/shuangpinprofile.h"
 #include <fcitx-utils/log.h>
 using namespace libime;
@@ -54,10 +55,22 @@ void checkProfile(const ShuangpinProfile &profile, bool hasSemicolon) {
     FCITX_ASSERT(validSyls.empty()) << validSyls;
 }
 
+void checkABC() {
+    ShuangpinProfile sp(ShuangpinBuiltinProfile::ABC);
+    auto iter = sp.table().find("an");
+    FCITX_ASSERT(iter != sp.table().end());
+    FCITX_ASSERT(iter->second.count(
+        PinyinSyllable(PinyinInitial::ZH, libime::PinyinFinal::UN)));
+    FCITX_ASSERT(!iter->second.count(
+        PinyinSyllable(PinyinInitial::Zero, libime::PinyinFinal::AN)));
+}
+
 void checkXiaoHe() {
-    PinyinEncoder::parseUserShuangpin(
-        "aiaaah", ShuangpinProfile(ShuangpinBuiltinProfile::Xiaohe),
-        PinyinFuzzyFlag::None)
+    ShuangpinProfile profile(ShuangpinBuiltinProfile::Xiaohe);
+    FCITX_ASSERT(profile.table().count("lv") == 1);
+    FCITX_ASSERT(profile.table().find("lv")->second.count(PinyinSyllable(
+                     PinyinInitial::L, libime::PinyinFinal::IU)) == 0);
+    PinyinEncoder::parseUserShuangpin("aiaaah", profile, PinyinFuzzyFlag::None)
         .dfs([](const SegmentGraphBase &segs, const std::vector<size_t> &path) {
             size_t s = 0;
             for (auto e : path) {
@@ -424,7 +437,7 @@ int main() {
     checkProfile(ShuangpinProfile(ShuangpinBuiltinProfile::PinyinJiajia),
                  false);
     checkProfile(ShuangpinProfile(ShuangpinBuiltinProfile::Xiaohe), false);
-
+    checkABC();
     checkXiaoHe();
 
     // wo jiu shi xiang ce shi yi xia
